// This code is derived from jcifs smb client library <jcifs at samba dot org>
// Ported by J. Arturo <webmaster at komodosoft dot net>
//  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
using System;
using System.IO;
using SharpCifs.Smb;
using SharpCifs.Util.Sharpen;
using Windows.Storage.Streams;
using System.Threading.Tasks;

namespace SharpCifs.Util.Transport
{
	/// <summary>
	/// This class simplifies communication for protocols that support
	/// multiplexing requests.
	/// </summary>
	/// <remarks>
	/// This class simplifies communication for protocols that support
	/// multiplexing requests. It encapsulates a stream and some protocol
	/// knowledge (provided by a concrete subclass) so that connecting,
	/// disconnecting, sending, and receiving can be syncronized
	/// properly. Apparatus is provided to send and receive requests
	/// concurrently.
	/// </remarks>
	public abstract class Transport : IRunnable
	{
        public AsyncLock ThisLock = new AsyncLock();
        
        internal static int Id;
        public static bool GlobalConnectionAborted = false;

		//internal static LogStream log = LogStream.GetInstance();

	    public LogStream Log
	    {
	        get
	        {
	            return LogStream.GetInstance();
	        }
	    }

		/// <exception cref="System.IO.IOException"></exception>
		public static async Task<int> Readn(DataReader inr, byte[] b, int off, int len)
		{
			//int i = 0;
			//int n = -5;
            await inr.LoadAsync((uint)len);
            len = (int)inr.UnconsumedBufferLength;
            if (len > 0)
            {
                byte[] data = new byte[len];
                inr.ReadBytes(data);
                Array.ConstrainedCopy(data, 0, b, off, len);
            }            
            /*while (i < len)
			{
                @in.Read(b, off + i, len - i);
				if (n <= 0)
				{
					break;
				}
				i += n;
			}
			return i;*/
            return len;
		}

		internal int State;

		internal string Name = "Transport" + Id++;

		internal Thread Thread;

		internal TransportException Te;

		protected internal Hashtable ResponseMap = new Hashtable();

		/// <exception cref="System.IO.IOException"></exception>
		protected internal abstract void MakeKey(ServerMessageBlock request);

		/// <exception cref="System.IO.IOException"></exception>
		protected internal abstract Task<ServerMessageBlock> PeekKey();

		/// <exception cref="System.IO.IOException"></exception>
        protected internal abstract Task DoSend(ServerMessageBlock request);

		/// <exception cref="System.IO.IOException"></exception>
        protected internal abstract Task DoRecv(Response response);

		/// <exception cref="System.IO.IOException"></exception>
        protected internal abstract Task DoSkip();

		/// <exception cref="System.IO.IOException"></exception>
        public virtual async Task Sendrecv(ServerMessageBlock request, Response response, long timeout, bool AlreadyLocked)
		{
            if (!AlreadyLocked)
            {
                using (var releaser = await ThisLock.LockAsync())
                {
                    await Sendrecv(request, response, timeout, true);
                    return;
                }
            }
			{
				MakeKey(request);
				response.IsReceived = false;
				try
				{
					ResponseMap.Put(request, response);
					await DoSend(request);
					response.Expiration = Runtime.CurrentTimeMillis() + timeout;
                    lock (this)
                    {
                        while (!response.IsReceived)
                        {
                            Runtime.Wait(this, timeout);
                            timeout = response.Expiration - Runtime.CurrentTimeMillis();
                            if (timeout <= 0 && !response.IsReceived)
                            {
                                throw new TransportException(Name + " timedout waiting for response to " + request
                                    );
                            }
                        }
                    }
				}
				catch (IOException ioe)
				{
					if (Log.Level > 2)
					{
						Runtime.PrintStackTrace(ioe, Log);
					}
					try
					{
						Disconnect(true,false);
					}
					catch (IOException ioe2)
					{
						Runtime.PrintStackTrace(ioe2, Log);
					}
					throw;
				}
				catch (Exception ie)
				{
					throw new TransportException(ie);
				}
				finally
				{
					//Sharpen.Collections.Remove(response_map, request);
                    ResponseMap.Remove(request);
				}
			}
		}

        private async Task Loop(Thread runThread)
		{
            while (Thread == runThread)
			{
				try
				{
					ServerMessageBlock key = await PeekKey();
					if (key == null)
					{
						throw new IOException("end of stream");
					}

                    Response response;
                    lock (this)
					{
						response = (Response)ResponseMap.Get(key);                      
                    }
                    {
						if (response == null)
						{
							if (Log.Level >= 4)
							{
								Log.WriteLine("Invalid key, skipping message");
							}
							await DoSkip();
						}
						else
						{
							await DoRecv(response);
							response.IsReceived = true;
                            lock (this)
							    Runtime.NotifyAll(this);
						}
					}
				}
				catch (Exception ex)
				{
					string msg = ex.Message;
					bool timeout = msg != null && msg.Equals("Read timed out");
					bool hard = timeout == false;
					if (!timeout && Log.Level >= 3)
					{
						Runtime.PrintStackTrace(ex, Log);
					}
					try
					{
                        Disconnect(hard, false);
					}
					catch (IOException ioe)
					{
						Runtime.PrintStackTrace(ioe, Log);
					}
				}
			}
		}

		/// <exception cref="System.Exception"></exception>
        protected internal abstract Task DoConnect(bool AlreadyLocked);

		/// <exception cref="System.IO.IOException"></exception>
		protected internal abstract Task DoDisconnect(bool hard, bool AlreadyLocked);

		/// <exception cref="SharpCifs.Util.Transport.TransportException"></exception>
		public virtual async Task Connect(long timeout)
		{
			//lock (this)
			{
				try
				{
					switch (State)
					{
						case 0:
						{
							break;
						}

						case 3:
						{
							return;
						}

						case 4:
						{
							// already connected
							State = 0;
							throw new TransportException("Connection in error", Te);
						}

						default:
						{
							//TransportException te = new TransportException("Invalid state: " + state);
							State = 0;
                            throw new TransportException("Invalid state: " + State);
						}
					}
					State = 1;
					Te = null;
					Thread = new Thread(this);
					Thread.SetDaemon(true);
					//lock (Thread)
					{
                        GlobalConnectionAborted = false;
						Thread.Start();
                        timeout += (System.DateTime.Now.Ticks / 10000);
                        while (System.DateTime.Now.Ticks / 10000 < timeout)
                        {
                            lock (Thread)
                                if (Runtime.Wait(Thread, 50))
                                    break;
                            await Task.Delay(50);  // yield, allow abort
                            if (GlobalConnectionAborted)
                                break;
                        }
						switch (State)
						{
							case 1:
							{
								State = 0;
								Thread = null;
								throw new TransportException("Connection timeout");
							}

							case 2:
							{
								if (Te != null)
								{
									State = 4;
									Thread = null;
									throw Te;
								}
								State = 3;
								return;
							}
						}
					}
				}
				catch (Exception ie)
				{
					State = 0;
					Thread = null;
					throw new TransportException(ie);
				}
				finally
				{
					if (State != 0 && State != 3 && State != 4)
					{
						if (Log.Level >= 1)
						{
							Log.WriteLine("Invalid state: " + State);
						}
						State = 0;
						Thread = null;
					}
				}
			}
		}

		/// <exception cref="System.IO.IOException"></exception>
        public virtual void Disconnect(bool hard, bool AlreadyLocked)
		{

            if (hard)
            {
                IOException ioe = null;
                switch (State)
                {
                    case 0:
                        {
                            return;
                        }

                    case 2:
                        {
                            hard = true;
                            goto case 3;
                        }

                    case 3:
                        {
                            if (ResponseMap.Count != 0 && !hard)
                            {
                                break;
                            }
                            try
                            {
                                DoDisconnect(hard, AlreadyLocked);
                            }
                            catch (IOException ioe0)
                            {
                                ioe = ioe0;
                            }
                            goto case 4;
                        }

                    case 4:
                        {
                            Thread = null;
                            State = 0;
                            break;
                        }

                    default:
                        {
                            if (Log.Level >= 1)
                            {
                                Log.WriteLine("Invalid state: " + State);
                            }
                            Thread = null;
                            State = 0;
                            break;
                        }
                }
                if (ioe != null)
                {
                    throw ioe;
                }

                return;
            }
            
            lock (this)
			{
				IOException ioe = null;
				switch (State)
				{
					case 0:
					{
						return;
					}

					case 2:
					{
						hard = true;
						goto case 3;
					}

					case 3:
					{
						if (ResponseMap.Count != 0 && !hard)
						{
							break;
						}
						try
						{
                            DoDisconnect(hard, AlreadyLocked);
						}
						catch (IOException ioe0)
						{
							ioe = ioe0;
						}
						goto case 4;
					}

					case 4:
					{
						Thread = null;
						State = 0;
						break;
					}

					default:
					{
						if (Log.Level >= 1)
						{
							Log.WriteLine("Invalid state: " + State);
						}
						Thread = null;
						State = 0;
						break;
					}
				}
				if (ioe != null)
				{
					throw ioe;
				}
			}
		}

		public virtual async void Run()
		{
			Thread runThread = Thread;
			Exception ex0 = null;
			try
			{
				await DoConnect(true);
			}
			catch (Exception ex)
			{
				ex0 = ex;
				// Defer to below where we're locked
				return;
			}
			finally
			{
				lock (runThread)
				{
					if (runThread != Thread)
					{
						if (ex0 != null)
						{
							if (Log.Level >= 2)
							{
								Runtime.PrintStackTrace(ex0, Log);
							}
						}
						//return;
					}
					if (ex0 != null)
					{
						Te = new TransportException(ex0);
					}
					State = 2;
					// run connected
					Runtime.Notify(runThread);
				}
			}
			await Loop(runThread);
		}

		public override string ToString()
		{
			return Name;
		}
	}
}
